"""
Functions to perform normal, weighted and robust fitting.
"""
from __future__ import annotations

import inspect
import warnings
from typing import Any, Callable

import numpy as np
import scipy
from geoutils.raster import subsample_array
from numpy.polynomial.polynomial import polyval, polyval2d

from xdem._typing import NDArrayf

try:
    from sklearn.linear_model import (
        HuberRegressor,
        LinearRegression,
        RANSACRegressor,
        TheilSenRegressor,
    )
    from sklearn.metrics import median_absolute_error
    from sklearn.pipeline import make_pipeline
    from sklearn.preprocessing import PolynomialFeatures

    _has_sklearn = True
except ImportError:
    _has_sklearn = False


def rmse(z: NDArrayf) -> float:
    """
    Return root mean square error
    :param z: Residuals between predicted and true value
    :return: Root Mean Square Error
    """
    return np.sqrt(np.nanmean(np.square(z)))


def huber_loss(z: NDArrayf) -> float:
    """
    Huber loss cost (reduces the weight of outliers)
    :param z: Residuals between predicted and true values
    :return: Huber cost
    """
    out = np.where(z > 1, 2 * np.sqrt(z[np.where(z > 1)]) - 1, np.square(z))

    return out.sum()


def soft_loss(z: NDArrayf, scale: float = 0.5) -> float:
    """
    Soft loss cost (reduces the weight of outliers)
    :param z: Residuals between predicted and true values
    :param scale: Scale factor
    :return: Soft loss cost
    """
    return np.sum(np.square(scale) * 2 * (np.sqrt(1 + np.square(z / scale)) - 1))


######################################################
# Most common functions for 1- or 2-D bias corrections
######################################################


def sumsin_1d(xx: NDArrayf, *params: NDArrayf) -> NDArrayf:
    """
    Sum of N sinusoids in 1D.

    :param xx: Array of coordinates.
    :param params: 3 x N parameters in order of amplitude (Y unit), wavelength (X unit) and phase (radians).
    """

    # Squeeze input in case it is a 1-D tuple or such
    xx = np.array(xx).squeeze()

    # Convert parameters to array
    p = np.array(params)

    # Indexes of amplitude, frequencies and phases
    aix = np.arange(0, len(p), 3)
    bix = np.arange(1, len(p), 3)
    cix = np.arange(2, len(p), 3)

    # Expand array to the same size as data.dim + 1, and move params to axis 0 for sum (ndmin moves it to last axis)
    p = np.moveaxis(np.array(p, ndmin=xx.ndim + 1), source=xx.ndim, destination=0)

    # Perform the sum of sinusoid
    val = np.sum(p[aix, :] * np.sin(2 * np.pi / p[bix, :] * np.expand_dims(xx, axis=0) + p[cix, :]), axis=0)

    return val


def polynomial_1d(xx: NDArrayf, *params: NDArrayf) -> NDArrayf:
    """
    N-order 1D polynomial.

    :param xx: 1D array of values.
    :param params: N polynomial parameters.

    :return: Output value.
    """
    return polyval(x=xx, c=params)


def polynomial_2d(xx: tuple[NDArrayf, NDArrayf], *params: NDArrayf) -> NDArrayf:
    """
    N-order 2D polynomial.

    :param xx: The two 1D array of values.
    :param params: The N parameters (a, b, c, etc.) of the polynomial.

    :returns: Output value.
    """

    # The number of parameters of np.polyval2d is order^2, so a square array needs to be passed
    poly_order = np.sqrt(len(params))

    if not poly_order.is_integer():
        raise ValueError(
            "The parameters of the 2D polynomial should have a length equal to order^2, "
            "see np.polyval2d for more details."
        )

    # We reshape the parameter into the N x N shape expected by NumPy
    c = np.array(params).reshape((int(poly_order), int(poly_order)))

    return polyval2d(x=xx[0], y=xx[1], c=c)


#######################################################################
# Convenience wrappers for robust N-order polynomial or sum of sin fits
#######################################################################


def _choice_best_order(cost: NDArrayf, margin_improvement: float = 20.0, verbose: bool = False) -> int:
    """
    Choice of the best order (polynomial, sum of sinusoids) with a margin of improvement. The best cost value does
    not necessarily mean the best predictive fit because high-degree polynomials tend to overfit, and sum of sinusoids
    as well. To mitigate this issue, we should choose the lesser order from which improvement becomes negligible.

    :param cost: cost function residuals to the polynomial
    :param margin_improvement: improvement margin (percentage) below which the lesser degree polynomial is kept
    :param verbose: if text should be printed

    :return: degree: degree for the best-fit polynomial
    """

    # Get percentage of spread from the minimal cost
    ind_min = cost.argmin()
    min_cost = cost[ind_min]
    perc_cost_improv = (cost - min_cost) / min_cost

    # Keep only good-performance costs that are within 20% of the minimal cost
    below_margin = np.logical_and(perc_cost_improv < margin_improvement / 100.0, np.arange(len(cost)) <= ind_min)

    # Choose the good-performance cost with lowest degree
    ind = next((i for i, j in enumerate(below_margin) if j))

    if verbose:
        print("Order " + str(ind_min + 1) + " has the minimum cost value of " + str(min_cost))
        print(
            "Order " + str(ind + 1) + " is selected as its cost is within a " + str(margin_improvement) + "% margin of"
            " the minimum cost"
        )

    return ind


def _wrapper_scipy_leastsquares(
    f: Callable[..., NDArrayf],
    xdata: NDArrayf,
    ydata: NDArrayf,
    sigma: NDArrayf | None = None,
    p0: NDArrayf = None,
    **kwargs: Any,
) -> tuple[float, NDArrayf]:
    """
    Wrapper function for scipy.optimize.least_squares: passes down keyword, extracts cost and final parameters, print
    statements in the console

    :param f: Function to fit.
    :param p0: Initial guess.
    :param x: X vector.
    :param y: Y vector.
    :param verbose: Whether to print out statements.
    :return:
    """

    # Get arguments of scipy.optimize.curve_fit and subfunction least_squares
    fun_args = scipy.optimize.curve_fit.__code__.co_varnames[: scipy.optimize.curve_fit.__code__.co_argcount]
    ls_args = scipy.optimize.least_squares.__code__.co_varnames[: scipy.optimize.least_squares.__code__.co_argcount]

    all_args = list(fun_args) + list(ls_args)

    # Check no other argument is left to be passed
    remaining_kwargs = kwargs.copy()
    for arg in all_args:
        remaining_kwargs.pop(arg, None)
    if len(remaining_kwargs) != 0:
        warnings.warn("Keyword arguments: " + ",".join(list(remaining_kwargs.keys())) + " were not used.")
    # Filter corresponding arguments before passing
    filtered_kwargs = {k: kwargs[k] for k in all_args if k in kwargs}

    # Run function with associated keyword arguments
    coefs = scipy.optimize.curve_fit(
        f=f,
        xdata=xdata,
        ydata=ydata,
        p0=p0,
        sigma=sigma,
        absolute_sigma=True,
        **filtered_kwargs,
    )[0]

    # Round results above the tolerance to get fixed results on different OS
    coefs = np.array([np.round(coef, 5) for coef in coefs])

    # If a specific loss function was passed, construct it to get the cost
    if "loss" in kwargs.keys():
        loss = kwargs["loss"]
        if "f_scale" in kwargs.keys():
            f_scale = kwargs["f_scale"]
        else:
            f_scale = 1.0
        from scipy.optimize._lsq.least_squares import construct_loss_function

        loss_func = construct_loss_function(m=ydata.size, loss=loss, f_scale=f_scale)
        cost = 0.5 * sum(np.atleast_1d(loss_func((f(xdata, *coefs) - ydata) ** 2, cost_only=True)))
    # Default is linear loss
    else:
        cost = 0.5 * sum((f(xdata, *coefs) - ydata) ** 2)

    return cost, coefs


def _wrapper_sklearn_robustlinear(
    model: PolynomialFeatures,
    cost_func: Callable[[NDArrayf, NDArrayf], float],
    xdata: NDArrayf,
    ydata: NDArrayf,
    sigma: NDArrayf | None = None,
    estimator_name: str = "Linear",
    **kwargs: Any,
) -> tuple[float, NDArrayf]:
    """
    Wrapper function of sklearn.linear_models: passes down keyword, extracts cost and final parameters, sets random
    states, scales input and de-scales output data, prints out statements

    :param model: Function model to fit (e.g., Polynomial features)
    :param cost_func: Cost function to use for optimization
    :param xdata: X vector
    :param ydata: Y vector
    :param estimator_name: Linear estimator to use (one of "Linear", "Theil-Sen", "RANSAC" and "Huber")
    :return:
    """
    # Select sklearn estimator
    dict_estimators = {
        "Linear": LinearRegression,
        "Theil-Sen": TheilSenRegressor,
        "RANSAC": RANSACRegressor,
        "Huber": HuberRegressor,
    }

    est = dict_estimators[estimator_name]

    # Get existing arguments of the sklearn estimator and model
    estimator_args = list(inspect.signature(est.__init__).parameters.keys())

    # Check no other argument is left to be passed
    remaining_kwargs = kwargs.copy()
    for arg in estimator_args:
        remaining_kwargs.pop(arg, None)
    if len(remaining_kwargs) != 0:
        warnings.warn("Keyword arguments: " + ",".join(list(remaining_kwargs.keys())) + " were not used.")
    # Filter corresponding arguments before passing
    filtered_kwargs = {k: kwargs[k] for k in estimator_args if k in kwargs}

    # TODO: Find out how to re-scale polynomial coefficient + doc on what is the best scaling for polynomials
    # # Scale output data (important for ML algorithms):
    # robust_scaler = RobustScaler().fit(x.reshape(-1,1))
    # x_scaled = robust_scaler.transform(x.reshape(-1,1))
    # # Fit scaled data
    # model.fit(x_scaled, y)
    # y_pred = model.predict(x_scaled)

    # Initialize estimator with arguments
    init_estimator = est(**filtered_kwargs)

    # Create pipeline
    pipeline = make_pipeline(model, init_estimator)

    # Run with data
    # The sample weight can only be passed if it exists in the estimator call
    if sigma is not None and "sample_weight" in inspect.signature(est.fit).parameters.keys():
        # The weight is the inverse of the squared standard error
        sample_weight = 1 / sigma**2
        # The argument name to pass it through a pipeline is "estimatorname__sample_weight"
        args = {est.__name__.lower() + "__sample_weight": sample_weight}
        pipeline.fit(xdata.reshape(-1, 1), ydata, *args)
    else:
        pipeline.fit(xdata.reshape(-1, 1), ydata)

    y_pred = pipeline.predict(xdata.reshape(-1, 1))

    # Calculate cost
    cost = cost_func(y_pred, ydata)

    # Get polynomial coefficients estimated with the estimators Linear, Theil-Sen and Huber
    if estimator_name in ["Linear", "Theil-Sen", "Huber"]:
        coefs = init_estimator.coef_
    # For some reason RANSAC doesn't store coef at the same place
    else:
        coefs = init_estimator.estimator_.coef_

    return cost, coefs


def robust_norder_polynomial_fit(
    xdata: NDArrayf,
    ydata: NDArrayf,
    sigma: NDArrayf | None = None,
    max_order: int = 6,
    estimator_name: str = "Theil-Sen",
    cost_func: Callable[[NDArrayf, NDArrayf], float] = median_absolute_error,
    margin_improvement: float = 20.0,
    subsample: float | int = 1,
    linear_pkg: str = "scipy",
    verbose: bool = False,
    random_state: None | np.random.RandomState | np.random.Generator | int = None,
    **kwargs: Any,
) -> tuple[NDArrayf, int]:
    """
    Given 1D vectors x and y, compute a robust polynomial fit to the data. Order is chosen automatically by comparing
    residuals for multiple fit orders of a given estimator.

    Any keyword argument will be passed down to scipy.optimize.least_squares and sklearn linear estimators.

    :param xdata: Input x data (N,).
    :param ydata: Input y data (N,).
    :param sigma: Standard error of y data (N,).
    :param max_order: Maximum polynomial order tried for the fit.
    :param estimator_name: robust estimator to use, one of 'Linear', 'Theil-Sen', 'RANSAC' or 'Huber'.
    :param cost_func: cost function taking as input two vectors y (true y), y' (predicted y) of same length.
    :param margin_improvement: improvement margin (percentage) below which the lesser degree polynomial is kept.
    :param subsample: If <= 1, will be considered a fraction of valid pixels to extract.
        If > 1 will be considered the number of pixels to extract.
    :param linear_pkg: package to use for Linear estimator, one of 'scipy' and 'sklearn'.
    :param random_state: Random seed.
    :param verbose: Whether to print text.

    :returns coefs, degree: Polynomial coefficients and degree for the best-fit polynomial
    """
    # Remove "f" and "absolute sigma" arguments passed, as both are fixed here
    if "f" in kwargs.keys():
        kwargs.pop("f")
    if "absolute_sigma" in kwargs.keys():
        kwargs.pop("absolute_sigma")

    # Raise errors for input string parameters
    if not isinstance(estimator_name, str) or estimator_name not in ["Linear", "Theil-Sen", "RANSAC", "Huber"]:
        raise ValueError('Attribute estimator must be one of "Linear", "Theil-Sen", "RANSAC" or "Huber".')
    if not isinstance(linear_pkg, str) or linear_pkg not in ["sklearn", "scipy"]:
        raise ValueError('Attribute linear_pkg must be one of "scipy" or "sklearn".')

    # Extract xdata from iterable
    if len(xdata) == 1:
        xdata = xdata[0]

    # Remove NaNs
    valid_data = np.logical_and(np.isfinite(ydata), np.isfinite(xdata))
    x = xdata[valid_data]
    y = ydata[valid_data]

    # Subsample data
    if subsample != 1:
        subsamp = subsample_array(x, subsample=subsample, return_indices=True, random_state=random_state)
        x = x[subsamp]
        y = y[subsamp]

    # Initialize cost function and output coefficients
    list_costs = np.empty(max_order)
    list_coeffs = np.zeros((max_order, max_order + 1))

    # Loop on polynomial degrees
    for deg in np.arange(1, max_order + 1):
        # If method is linear and package scipy
        if estimator_name == "Linear" and linear_pkg == "scipy":

            # Define the initial guess
            p0 = np.polyfit(x, y, deg)

            # Run the linear method with scipy
            try:
                cost, coef = _wrapper_scipy_leastsquares(
                    f=polynomial_1d, xdata=x, ydata=y, p0=p0, sigma=sigma, **kwargs
                )
            except RuntimeError:
                cost = np.inf
                coef = np.array([np.nan for i in range(len(p0))])

        else:
            # Otherwise, we use sklearn
            if not _has_sklearn:
                raise ValueError("Optional dependency needed. Install 'scikit-learn'")

            # Define the polynomial model to insert in the pipeline
            model = PolynomialFeatures(degree=deg)

            # Run the linear method with sklearn
            cost, coef = _wrapper_sklearn_robustlinear(
                model, estimator_name=estimator_name, cost_func=cost_func, xdata=x, ydata=y, sigma=sigma, **kwargs
            )

        list_costs[deg - 1] = cost
        list_coeffs[deg - 1, 0 : coef.size] = coef

    # Choose the best polynomial with a margin of improvement on the cost
    final_index = _choice_best_order(cost=list_costs, margin_improvement=margin_improvement, verbose=verbose)

    # The degree of the best polynomial corresponds to the index plus one
    return np.trim_zeros(list_coeffs[final_index], "b"), final_index + 1


def _cost_sumofsin(
    x: NDArrayf,
    y: NDArrayf,
    cost_func: Callable[[NDArrayf], float],
    *p: NDArrayf,
) -> float:
    """
    Calculate robust cost function for sum of sinusoids
    """
    z = y - sumsin_1d(x, *p)
    return cost_func(z)


def robust_nfreq_sumsin_fit(
    xdata: NDArrayf,
    ydata: NDArrayf,
    sigma: NDArrayf | None = None,
    max_nb_frequency: int = 3,
    bounds_amp_wave_phase: list[tuple[float, float]] | None = None,
    cost_func: Callable[[NDArrayf], float] = soft_loss,
    subsample: float | int = 1,
    hop_length: float | None = None,
    random_state: None | np.random.RandomState | np.random.Generator | int = None,
    verbose: bool = False,
    **kwargs: Any,
) -> tuple[NDArrayf, int]:
    """
    Given 1D vectors x and y, compute a robust sum of sinusoid fit to the data. The number of frequency is chosen
    automatically by comparing residuals for multiple fit orders of a given estimator.

    Any keyword argument will be passed down to scipy.optimize.basinhopping.

    :param xdata: Input x data (N,).
    :param ydata: Input y data (N,).
    :param sigma: Standard error of y data (N,).
    :param max_nb_frequency: Maximum number of sinusoid of different frequencies.
    :param bounds_amp_wave_phase: Bounds for amplitude, wavelength and phase (L, 3, 2) and
        with mean value used for initialization.
    :param hop_length: Jump in function values to optimize basinhopping algorithm search (for best results, should be
    comparable to the separation in function value between local minima).
    :param cost_func: Cost function taking as input two vectors y (true y), y' (predicted y) of same length.
    :param subsample: If <= 1, will be considered a fraction of valid pixels to extract.
        If > 1 will be considered the number of pixels to extract.
    :param random_state: Random seed.
    :param verbose: If text should be printed.
    :param kwargs: Keyword arguments to pass to scipy.optimize.basinhopping

    :returns coefs, degree: sinusoid coefficients (amplitude, frequency, phase) x N, Number N of summed sinusoids
    """

    # Remove "f" and "absolute sigma" arguments passed, as both are fixed here
    if "f" in kwargs.keys():
        kwargs.pop("f")
    if "absolute_sigma" in kwargs.keys():
        kwargs.pop("absolute_sigma")

    # Extract xdata from iterable
    if len(xdata) == 1:
        xdata = xdata[0]

    # Check if there is a number of iterations to stop the run if the global minimum candidate remains the same.
    if "niter_success" not in kwargs.keys():
        # Check if there is a number of basin-hopping iterations passed down to the function.
        if "niter" not in kwargs.keys():
            niter_success = 40
        else:
            niter_success = min(40, kwargs.get("niter"))  # type: ignore

        kwargs.update({"niter_success": niter_success})

    def wrapper_cost_sumofsin(p: NDArrayf, x: NDArrayf, y: NDArrayf) -> float:
        return _cost_sumofsin(x, y, cost_func, *p)

    # If no significant resolution is provided, assume that it is the mean difference between sampled X values
    if hop_length is None:
        x_res = np.mean(np.diff(np.sort(xdata)))
        hop_length = x_res

    # Loop on all frequencies
    costs = np.empty(max_nb_frequency)
    amp_freq_phase = np.zeros((max_nb_frequency, 3 * max_nb_frequency)) * np.nan

    for nb_freq in np.arange(1, max_nb_frequency + 1):

        if verbose:
            print(f"Fitting with {nb_freq} frequency")

        b = bounds_amp_wave_phase
        # If bounds are not provided, define as the largest possible bounds
        if b is None:
            # For the amplitude, from Y values
            lb_amp = 0
            ub_amp = ydata.max() - ydata.min()
            # For phase: all possible values for a sinusoid
            lb_phase = 0
            ub_phase = 2 * np.pi
            # For the wavelength: from the resolution and coordinate extent
            # (we don't want the lower bound to be zero, to avoid divisions by zero)
            lb_wavelength = hop_length / 5
            ub_wavelength = xdata.max() - xdata.min()

            b = []
            for _i in range(nb_freq):
                b += [(lb_amp, ub_amp), (lb_wavelength, ub_wavelength), (lb_phase, ub_phase)]

        # Format lower and upper bounds for scipy
        lb = np.asarray([b[i][0] for i in range(3 * nb_freq)])
        ub = np.asarray([b[i][1] for i in range(3 * nb_freq)])
        # Insert in a scipy bounds object
        scipy_bounds = scipy.optimize.Bounds(lb, ub)
        # First guess for the mean parameters
        p0 = np.divide(lb + ub, 2).squeeze()

        if verbose:
            print("Bounds")
            print(lb)
            print(ub)

        # Minimize the globalization with a larger number of points
        minimizer_kwargs = dict(args=(xdata, ydata), method="L-BFGS-B", bounds=scipy_bounds)
        myresults = scipy.optimize.basinhopping(
            wrapper_cost_sumofsin,
            p0,
            disp=verbose,
            T=hop_length * 50,
            minimizer_kwargs=minimizer_kwargs,
            seed=random_state,
            **kwargs,
        )
        myresults = myresults.lowest_optimization_result
        myresults_x = np.array([np.round(myres, 5) for myres in myresults.x])

        if verbose:
            print("Final result")
            print(myresults_x)

        # Write results for this number of frequency
        costs[nb_freq - 1] = wrapper_cost_sumofsin(myresults_x, xdata, ydata)
        amp_freq_phase[nb_freq - 1, 0 : 3 * nb_freq] = myresults_x

    # Replace NaN cost by infinity
    costs[np.isnan(costs)] = np.inf

    if verbose:
        print("Costs")
        print(costs)

    final_index = _choice_best_order(cost=costs)

    final_coefs = amp_freq_phase[final_index][~np.isnan(amp_freq_phase[final_index])]

    if verbose:
        print("Selecting best performing number of frequencies:")
        print(final_coefs)

    # If an amplitude coefficient is almost zero, remove the coefs of that frequency and lower the degree
    final_degree = final_index + 1
    for i in range(final_index + 1):
        # If an amplitude has an estimated value of less than 0.1% the signal bounds (percentiles for robustness)
        # And if the degree is higher than 2 (need at least degree 1 return)
        if (
            np.abs(final_coefs[3 * i]) < (np.nanpercentile(ydata, 90) - np.nanpercentile(ydata, 10)) / 1000
            and len(final_coefs) > 3
        ):
            final_coefs = np.delete(final_coefs, slice(3 * i, 3 * i + 3))
            final_degree -= 1
            break

    # Re-order frequencies by highest amplitude
    amplitudes = final_coefs[0::3]
    indices = np.flip(np.argsort(amplitudes))
    new_amplitudes = amplitudes[indices]
    new_wavelengths = final_coefs[1::3][indices]
    new_phases = final_coefs[2::3][indices]

    final_coefs = np.array(
        [(new_amplitudes[i], new_wavelengths[i], new_phases[i]) for i in range(final_degree)]
    ).flatten()

    # The number of frequencies corresponds to the final index plus one
    return final_coefs, final_degree



"""Functions to test the filtering tools."""
from __future__ import annotations

import geoutils as gu
import numpy as np
import pytest

import xdem


class TestFilters:
    """Test cases for the filter functions."""

    # Load example data.
    dem_2009 = gu.Raster(xdem.examples.get_path("longyearbyen_ref_dem"))
    dem_1990 = gu.Raster(xdem.examples.get_path("longyearbyen_tba_dem")).reproject(dem_2009, silent=True)

    def test_gauss(self) -> None:
        """Test applying the various Gaussian filters on DEMs with/without NaNs"""

        # Test applying scipy's Gaussian filter
        # smoothing should not yield values below.above original DEM
        dem_array = self.dem_1990.data
        dem_sm = xdem.filters.gaussian_filter_scipy(dem_array, sigma=5)
        assert np.min(dem_array) < np.min(dem_sm)
        assert np.max(dem_array) > np.max(dem_sm)
        assert dem_array.shape == dem_sm.shape

        # Test applying OpenCV's Gaussian filter
        dem_sm2 = xdem.filters.gaussian_filter_cv(dem_array, sigma=5)
        assert np.min(dem_array) < np.min(dem_sm2)
        assert np.max(dem_array) > np.max(dem_sm2)
        assert dem_array.shape == dem_sm2.shape

        # Assert that both implementations yield similar results
        assert np.nanmax(np.abs(dem_sm - dem_sm2)) < 1e-3

        # Test that it works with NaNs too
        nan_count = 1000
        cols = np.random.randint(0, high=self.dem_1990.width - 1, size=nan_count, dtype=int)
        rows = np.random.randint(0, high=self.dem_1990.height - 1, size=nan_count, dtype=int)
        dem_with_nans = np.copy(self.dem_1990.data).squeeze()
        dem_with_nans[rows, cols] = np.nan

        dem_sm = xdem.filters.gaussian_filter_scipy(dem_with_nans, sigma=10)
        assert np.nanmin(dem_with_nans) < np.min(dem_sm)
        assert np.nanmax(dem_with_nans) > np.max(dem_sm)

        dem_sm = xdem.filters.gaussian_filter_cv(dem_with_nans, sigma=10)
        assert np.nanmin(dem_with_nans) < np.min(dem_sm)
        assert np.nanmax(dem_with_nans) > np.max(dem_sm)

        # Test that it works with 3D arrays
        array_3d = np.vstack((dem_array[np.newaxis, :], dem_array[np.newaxis, :]))
        dem_sm = xdem.filters.gaussian_filter_scipy(array_3d, sigma=5)
        assert array_3d.shape == dem_sm.shape

        # 3D case not implemented with OpenCV
        pytest.raises(NotImplementedError, xdem.filters.gaussian_filter_cv, array_3d, sigma=5)

        # Tests that it fails with 1D arrays with appropriate error
        data = dem_array[:, 0]
        pytest.raises(ValueError, xdem.filters.gaussian_filter_scipy, data, sigma=5)
        pytest.raises(ValueError, xdem.filters.gaussian_filter_cv, data, sigma=5)

    def test_dist_filter(self) -> None:
        """Test that distance_filter works"""

        # Calculate dDEM
        ddem = self.dem_2009.data - self.dem_1990.data

        # Add random outliers
        count = 1000
        cols = np.random.randint(0, high=self.dem_1990.width - 1, size=count, dtype=int)
        rows = np.random.randint(0, high=self.dem_1990.height - 1, size=count, dtype=int)
        ddem.data[rows, cols] = 5000

        # Filter gross outliers
        filtered_ddem = xdem.filters.distance_filter(ddem.data, radius=20, outlier_threshold=50)

        # Check that all outliers were properly filtered
        assert np.all(np.isnan(filtered_ddem[rows, cols]))

        # Assert that non filtered pixels remain the same
        assert ddem.data.shape == filtered_ddem.shape
        assert np.all(ddem.data[np.isfinite(filtered_ddem)] == filtered_ddem[np.isfinite(filtered_ddem)])

        # Check that it works with NaNs too
        ddem.data[rows[:500], cols[:500]] = np.nan
        filtered_ddem = xdem.filters.distance_filter(ddem.data, radius=20, outlier_threshold=50)
        assert np.all(np.isnan(filtered_ddem[rows, cols]))



""" Functions to test the DEM tools."""
from __future__ import annotations

import os
import tempfile
import warnings
from typing import Any

import geoutils.raster as gr
import geoutils.raster.satimg as si
import numpy as np
import pytest
import rasterio as rio
from geoutils.raster.raster import _default_rio_attrs
from pyproj import CRS
from pyproj.transformer import Transformer

import xdem
import xdem.vcrs
from xdem.dem import DEM

DO_PLOT = False


class TestDEM:
    def test_init(self) -> None:
        """Test that inputs work properly in DEM class init."""
        fn_img = xdem.examples.get_path("longyearbyen_ref_dem")

        # From filename
        dem = DEM(fn_img)
        assert isinstance(dem, DEM)

        # From DEM
        dem2 = DEM(dem)
        assert isinstance(dem2, DEM)

        # From Raster
        r = gr.Raster(fn_img)
        dem3 = DEM(r)
        assert isinstance(dem3, DEM)

        # From SatelliteImage
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Parse metadata from file not implemented")
            img = si.SatelliteImage(fn_img)
        dem4 = DEM(img)
        assert isinstance(dem4, DEM)

        list_dem = [dem, dem2, dem3, dem4]

        # Check all attributes
        attrs = [at for at in _default_rio_attrs if at not in ["name", "dataset_mask", "driver"]]
        all_attrs = attrs + si.satimg_attrs + xdem.dem.dem_attrs
        for attr in all_attrs:
            attrs_per_dem = [idem.__getattribute__(attr) for idem in list_dem]
            assert all(at == attrs_per_dem[0] for at in attrs_per_dem)

        assert np.logical_and.reduce(
            (
                np.array_equal(dem.data, dem2.data, equal_nan=True),
                np.array_equal(dem2.data, dem3.data, equal_nan=True),
                np.array_equal(dem3.data, dem4.data, equal_nan=True),
            )
        )

        assert np.logical_and.reduce(
            (
                np.all(dem.data.mask == dem2.data.mask),
                np.all(dem2.data.mask == dem3.data.mask),
                np.all(dem3.data.mask == dem4.data.mask),
            )
        )

        # Check that an error is raised when more than one band is provided
        with pytest.raises(ValueError, match="DEM rasters should be composed of one band only"):
            xdem.DEM.from_array(
                data=np.zeros(shape=(2, 5, 5)),
                transform=rio.transform.from_bounds(0, 0, 1, 1, 5, 5),
                crs=None,
                nodata=None,
            )

    def test_init__vcrs(self) -> None:
        """Test that vcrs is set properly during instantiation."""

        # Tests 1: instantiation with a file that has a 2D CRS

        # First, check a DEM that does not have any vertical CRS set
        fn_img = xdem.examples.get_path("longyearbyen_ref_dem")
        dem = DEM(fn_img)
        assert dem.vcrs is None

        # Setting a vertical CRS during instantiation should work here
        dem = DEM(fn_img, vcrs="EGM96")
        assert dem.vcrs_name == "EGM96 height"

        # Tests 2: instantiation with a file that has a 3D CRS
        # Create such a file
        dem = DEM(fn_img)
        dem_reproj = dem.reproject(dst_crs=4979)

        # Save to temporary folder
        temp_dir = tempfile.TemporaryDirectory()
        temp_file = os.path.join(temp_dir.name, "test.tif")
        dem_reproj.save(temp_file)

        # Check opening a DEM with a 3D CRS sets the vcrs
        dem_3d = DEM(temp_file)
        assert dem_3d.vcrs == "Ellipsoid"

        # Check that a warning is raised when trying to override with user input
        with pytest.warns(
            UserWarning,
            match="The CRS in the raster metadata already has a vertical component, "
            "the user-input 'EGM08' will override it.",
        ):
            DEM(temp_file, vcrs="EGM08")

    def test_from_array(self) -> None:
        """Test that overridden from_array works as expected."""

        # Create a 5x5 DEM
        data = np.ones((5, 5))
        transform = rio.transform.from_bounds(0, 0, 1, 1, 5, 5)
        crs = CRS("EPSG:4326")
        nodata = -9999
        vcrs = "EGM08"
        dem = DEM.from_array(data=data, transform=transform, crs=crs, nodata=nodata, vcrs=vcrs)

        # Check output matches
        assert isinstance(dem, DEM)
        assert isinstance(dem.data, np.ma.masked_array)
        assert np.array_equal(dem.data.data, np.ones((5, 5)))
        assert dem.transform == transform
        assert dem.crs == crs
        assert dem.nodata == nodata
        assert dem.vcrs == xdem.vcrs._vcrs_from_user_input(vcrs_input=vcrs)

    def test_from_array__vcrs(self) -> None:
        """Test that overridden from_array rightly sets the vertical CRS."""

        # Create a 5x5 DEM with a 2D CRS
        transform = rio.transform.from_bounds(0, 0, 1, 1, 5, 5)
        dem = DEM.from_array(data=np.ones((5, 5)), transform=transform, crs=CRS("EPSG:4326"), nodata=None, vcrs=None)
        assert dem.vcrs is None

        # One with a 3D ellipsoid CRS
        dem = DEM.from_array(data=np.ones((5, 5)), transform=transform, crs=CRS("EPSG:4979"), nodata=None, vcrs=None)
        assert dem.vcrs == "Ellipsoid"

        # One with a 2D and the ellipsoid vertical CRS
        dem = DEM.from_array(
            data=np.ones((5, 5)), transform=transform, crs=CRS("EPSG:4326"), nodata=None, vcrs="Ellipsoid"
        )
        assert dem.vcrs == "Ellipsoid"

        # One with a compound CRS
        dem = DEM.from_array(
            data=np.ones((5, 5)), transform=transform, crs=CRS("EPSG:4326+5773"), nodata=None, vcrs=None
        )
        assert dem.vcrs == CRS("EPSG:5773")

        # One with a CRS and vertical CRS
        dem = DEM.from_array(
            data=np.ones((5, 5)), transform=transform, crs=CRS("EPSG:4326"), nodata=None, vcrs=CRS("EPSG:5773")
        )
        assert dem.vcrs == CRS("EPSG:5773")

    def test_copy(self) -> None:
        """
        Test that the copy method works as expected for DEM. In particular
        when copying r to r2:
        - if r.data is modified and r copied, the updated data is copied
        - if r is copied, r.data changed, r2.data should be unchanged
        """
        # Open dataset, update data and make a copy
        r = xdem.dem.DEM(xdem.examples.get_path("longyearbyen_ref_dem"))
        r.data += 5
        r2 = r.copy()

        # Objects should be different (not pointing to the same memory)
        assert r is not r2

        # Check the object is a DEM
        assert isinstance(r2, xdem.dem.DEM)

        # Check all immutable attributes are equal
        # raster_attrs = ['bounds', 'count', 'crs', 'dtypes', 'height', 'indexes', 'nodata',
        #                    'res', 'shape', 'transform', 'width']
        # satimg_attrs = ['satellite', 'sensor', 'product', 'version', 'tile_name', 'datetime']
        # dem_attrs = ['vcrs', 'vcrs_grid', 'vcrs_name', 'ccrs']

        # using list directly available in Class
        attrs = [at for at in _default_rio_attrs if at not in ["name", "dataset_mask", "driver"]]
        all_attrs = attrs + si.satimg_attrs + xdem.dem.dem_attrs
        for attr in all_attrs:
            assert r.__getattribute__(attr) == r2.__getattribute__(attr)

        # Check data array
        assert np.array_equal(r.data, r2.data, equal_nan=True)

        # Check dataset_mask array
        assert np.all(r.data.mask == r2.data.mask)

        # Check that if r.data is modified, it does not affect r2.data
        r.data += 5
        assert not np.array_equal(r.data, r2.data, equal_nan=True)

        # Check that the new_array argument indeed modifies the raster
        r3 = r.copy(new_array=r2.data)

        assert np.array_equal(r3.data, r2.data)

    def test_set_vcrs(self) -> None:
        """Tests to set the vertical CRS."""

        fn_dem = xdem.examples.get_path("longyearbyen_ref_dem")
        dem = DEM(fn_dem)

        # -- Test 1: we check with names --

        # Check setting ellipsoid
        dem.set_vcrs(new_vcrs="Ellipsoid")
        assert dem.vcrs_name is not None
        assert "Ellipsoid (No vertical CRS)." in dem.vcrs_name
        assert dem.vcrs_grid is None

        # Check setting EGM96
        dem.set_vcrs(new_vcrs="EGM96")
        assert dem.vcrs_name == "EGM96 height"
        assert dem.vcrs_grid == "us_nga_egm96_15.tif"

        # Check setting EGM08
        dem.set_vcrs(new_vcrs="EGM08")
        assert dem.vcrs_name == "EGM2008 height"
        assert dem.vcrs_grid == "us_nga_egm08_25.tif"

        # -- Test 2: we check with grids --
        dem.set_vcrs(new_vcrs="us_nga_egm96_15.tif")
        assert dem.vcrs_name == "unknown using geoidgrids=us_nga_egm96_15.tif"
        assert dem.vcrs_grid == "us_nga_egm96_15.tif"

        dem.set_vcrs(new_vcrs="us_nga_egm08_25.tif")
        assert dem.vcrs_name == "unknown using geoidgrids=us_nga_egm08_25.tif"
        assert dem.vcrs_grid == "us_nga_egm08_25.tif"

        # Check that other existing grids are well detected in the pyproj.datadir
        dem.set_vcrs(new_vcrs="is_lmi_Icegeoid_ISN93.tif")

        # Check that non-existing grids raise errors
        with pytest.raises(
            ValueError,
            match="The provided grid 'the best grid' does not exist at https://cdn.proj.org/. "
            "Provide an existing grid.",
        ):
            dem.set_vcrs(new_vcrs="the best grid")

    def test_to_vcrs(self) -> None:
        """Tests the conversion of vertical CRS."""

        fn_dem = xdem.examples.get_path("longyearbyen_ref_dem")
        dem = DEM(fn_dem)

        # Reproject in WGS84 2D
        dem = dem.reproject(dst_crs=4326)
        dem_before_trans = dem.copy()

        # Set ellipsoid as vertical reference
        dem.set_vcrs(new_vcrs="Ellipsoid")
        ccrs_init = dem.ccrs
        median_before = np.nanmean(dem)
        # Transform to EGM96 geoid
        dem.to_vcrs(dst_vcrs="EGM96")
        median_after = np.nanmean(dem)

        # About 32 meters of difference in Svalbard between EGM96 geoid and ellipsoid
        assert median_after - median_before == pytest.approx(-32, rel=0.1)

        # Check that the results are consistent with the operation done independently
        ccrs_dest = xdem.vcrs._build_ccrs_from_crs_and_vcrs(dem.crs, xdem.vcrs._vcrs_from_user_input("EGM96"))
        transformer = Transformer.from_crs(crs_from=ccrs_init, crs_to=ccrs_dest, always_xy=True)

        xx, yy = dem.coords()
        x = xx[5, 5]
        y = yy[5, 5]
        z = dem_before_trans.data[5, 5]
        z_out = transformer.transform(xx=x, yy=y, zz=z)[2]

        assert z_out == pytest.approx(dem.data[5, 5])

    def test_to_vcrs__equal_warning(self) -> None:
        """Test that DEM.to_vcrs() does not transform if both 3D CRS are equal."""

        fn_dem = xdem.examples.get_path("longyearbyen_ref_dem")
        dem = DEM(fn_dem)

        # With both inputs as names
        dem.set_vcrs("EGM96")
        with pytest.warns(
            UserWarning, match="Source and destination vertical CRS are the same, " "skipping vertical transformation."
        ):
            dem.to_vcrs("EGM96")

        # With one input as name, the other as CRS
        dem.set_vcrs("Ellipsoid")
        with pytest.warns(
            UserWarning, match="Source and destination vertical CRS are the same, " "skipping vertical transformation."
        ):
            dem.to_vcrs(CRS("EPSG:4979"))

    # Compare to manually-extracted shifts at specific coordinates for the geoid grids
    egm96_chile = {"grid": "us_nga_egm96_15.tif", "lon": -68, "lat": -20, "shift": 42}
    egm08_chile = {"grid": "us_nga_egm08_25.tif", "lon": -68, "lat": -20, "shift": 42}
    geoid96_alaska = {"grid": "us_noaa_geoid06_ak.tif", "lon": -145, "lat": 62, "shift": 17}
    isn93_iceland = {"grid": "is_lmi_Icegeoid_ISN93.tif", "lon": -18, "lat": 65, "shift": 68}

    @pytest.mark.parametrize("grid_shifts", [egm08_chile, egm08_chile, geoid96_alaska, isn93_iceland])  # type: ignore
    def test_to_vcrs__grids(self, grid_shifts: dict[str, Any]) -> None:
        """Tests grids to convert vertical CRS."""

        # Using an arbitrary elevation of 100 m (no influence on the transformation)
        dem = DEM.from_array(
            data=np.array([[100]]),
            transform=rio.transform.from_bounds(
                grid_shifts["lon"], grid_shifts["lat"], grid_shifts["lon"] + 0.01, grid_shifts["lat"] + 0.01, 0.01, 0.01
            ),
            crs=CRS.from_epsg(4326),
            nodata=None,
        )
        dem.set_vcrs("Ellipsoid")

        # Transform to the vertical CRS of the grid
        dem.to_vcrs(grid_shifts["grid"])

        # Compare the elevation difference
        z_diff = 100 - dem.data[0, 0]

        # Check the shift is the one expect within 10%
        assert z_diff == pytest.approx(grid_shifts["shift"], rel=0.1)



"""Utility functions to download and find example data."""
import os
import tarfile
import tempfile
import urllib.request
from distutils.dir_util import copy_tree

import geoutils as gu

import xdem

_EXAMPLES_DIRECTORY = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "examples", "data"))
# Absolute filepaths to the example files.
_FILEPATHS_DATA = {
    "longyearbyen_ref_dem": os.path.join(_EXAMPLES_DIRECTORY, "Longyearbyen", "data", "DEM_2009_ref.tif"),
    "longyearbyen_tba_dem": os.path.join(_EXAMPLES_DIRECTORY, "Longyearbyen", "data", "DEM_1990.tif"),
    "longyearbyen_glacier_outlines": os.path.join(
        _EXAMPLES_DIRECTORY, "Longyearbyen", "data", "glacier_mask", "CryoClim_GAO_SJ_1990.shp"
    ),
    "longyearbyen_glacier_outlines_2010": os.path.join(
        _EXAMPLES_DIRECTORY, "Longyearbyen", "data", "glacier_mask", "CryoClim_GAO_SJ_2010.shp"
    ),
}

_FILEPATHS_PROCESSED = {
    "longyearbyen_ddem": os.path.join(_EXAMPLES_DIRECTORY, "Longyearbyen", "processed", "dDEM_2009_minus_1990.tif"),
    "longyearbyen_tba_dem_coreg": os.path.join(_EXAMPLES_DIRECTORY, "Longyearbyen", "processed", "DEM_1990_coreg.tif"),
}

available = list(_FILEPATHS_DATA.keys()) + list(_FILEPATHS_PROCESSED.keys())


def download_longyearbyen_examples(overwrite: bool = False) -> None:
    """
    Fetch the Longyearbyen example files.

    :param overwrite: Do not download the files again if they already exist.
    """
    if not overwrite and all(map(os.path.isfile, list(_FILEPATHS_DATA.values()))):
        # print("Datasets exist")
        return

    # If we ask for overwrite, also remove the processed test data
    if overwrite:
        for fn in list(_FILEPATHS_PROCESSED.values()):
            if os.path.exists(fn):
                os.remove(fn)

    # Static commit hash to be bumped every time it needs to be.
    commit = "fd832bc2e366cf2ba8b543f7e43f90ee02384f4f"
    # The URL from which to download the repository
    url = f"https://github.com/GlacioHack/xdem-data/tarball/main#commit={commit}"

    # Create a temporary directory to extract the tarball in.
    temp_dir = tempfile.TemporaryDirectory()
    tar_path = os.path.join(temp_dir.name, "data.tar.gz")

    response = urllib.request.urlopen(url)
    # If the response was right, download the tarball to the temporary directory
    if response.getcode() == 200:
        with open(tar_path, "wb") as outfile:
            outfile.write(response.read())
    else:
        raise ValueError(f"Longyearbyen data fetch gave non-200 response: {response.status_code}")

    # Extract the tarball
    with tarfile.open(tar_path) as tar:
        tar.extractall(temp_dir.name)

    # Find the first directory in the temp_dir (should only be one) and construct the Longyearbyen data dir path.
    dir_name = os.path.join(
        temp_dir.name,
        [dirname for dirname in os.listdir(temp_dir.name) if os.path.isdir(os.path.join(temp_dir.name, dirname))][0],
        "data",
        "Longyearbyen",
    )

    # Copy the data to the examples directory.
    copy_tree(dir_name, os.path.join(_EXAMPLES_DIRECTORY, "Longyearbyen", "data"))


def process_coregistered_examples(name: str, overwrite: bool = False) -> None:
    """
    Process the Longyearbyen example files into a dDEM (to avoid repeating this in many test/documentation steps).

    :param name: Name of test data
    :param overwrite: Do not download the files again if they already exist.
    """

    # If the file called already exists and overwrite is False, do nothing
    if not overwrite and os.path.isfile(_FILEPATHS_PROCESSED[name]):
        return

    # Check that data is downloaded before attempting processing
    download_longyearbyen_examples(overwrite=False)

    # If the ddem file does not exist, create it
    if not os.path.isfile(_FILEPATHS_PROCESSED["longyearbyen_ddem"]):
        reference_raster = gu.Raster(_FILEPATHS_DATA["longyearbyen_ref_dem"])
        to_be_aligned_raster = gu.Raster(_FILEPATHS_DATA["longyearbyen_tba_dem"])
        glacier_mask = gu.Vector(_FILEPATHS_DATA["longyearbyen_glacier_outlines"])
        inlier_mask = ~glacier_mask.create_mask(reference_raster)

        nuth_kaab = xdem.coreg.NuthKaab()
        nuth_kaab.fit(reference_raster, to_be_aligned_raster, inlier_mask=inlier_mask)
        aligned_raster = nuth_kaab.apply(to_be_aligned_raster, resample=True)

        diff = reference_raster - aligned_raster

        # Save it so that future calls won't need to recreate the file
        os.makedirs(os.path.dirname(_FILEPATHS_PROCESSED["longyearbyen_ddem"]), exist_ok=True)
        diff.save(_FILEPATHS_PROCESSED["longyearbyen_ddem"])

    # If the tba_dem_coreg file does not exist, create it
    if not os.path.isfile(_FILEPATHS_PROCESSED["longyearbyen_tba_dem_coreg"]):

        dem_2009 = xdem.DEM(get_path("longyearbyen_ref_dem"), silent=True)
        ddem = xdem.DEM(get_path("longyearbyen_ddem"), silent=True)

        # Save it so that future calls won't need to recreate the file
        (dem_2009 - ddem).save(_FILEPATHS_PROCESSED["longyearbyen_tba_dem_coreg"])


def get_path(name: str) -> str:
    """
    Get path of example data. List of available files can be found in "examples.available".
    :param name: Name of test data
    :return:
    """

    if name in list(_FILEPATHS_DATA.keys()):
        download_longyearbyen_examples()
        return _FILEPATHS_DATA[name]
    elif name in list(_FILEPATHS_PROCESSED.keys()):
        process_coregistered_examples(name)
        return _FILEPATHS_PROCESSED[name]
    else:
        raise ValueError(
            'Data name should be one of "'
            + '" , "'.join(list(_FILEPATHS_DATA.keys()) + list(_FILEPATHS_PROCESSED.keys()))
            + '".'
        )



"""
Functions to test the fitting tools.
"""
import platform
import warnings

import numpy as np
import pytest
from sklearn.metrics import mean_squared_error, median_absolute_error

import xdem


class TestRobustFitting:
    @pytest.mark.parametrize(
        "pkg_estimator",
        [
            ("sklearn", "Linear"),
            ("scipy", "Linear"),
            ("sklearn", "Theil-Sen"),
            ("sklearn", "RANSAC"),
            ("sklearn", "Huber"),
        ],
    )  # type: ignore
    def test_robust_norder_polynomial_fit(self, pkg_estimator: str) -> None:

        # Define x vector
        x = np.linspace(1, 10, 1000)
        # Define exact polynomial
        true_coefs = [-100, 5, 3, 2]
        y = np.polyval(np.flip(true_coefs), x).astype(np.float32)

        # Run fit
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="lbfgs failed to converge")
            coefs, deg = xdem.fit.robust_norder_polynomial_fit(
                x,
                y,
                linear_pkg=pkg_estimator[0],
                estimator_name=pkg_estimator[1],
                random_state=42,
                margin_improvement=50,
            )

        # Check coefficients are constrained
        assert deg == 3 or deg == 4
        error_margins = [100, 5, 2, 1]
        for i in range(4):
            assert coefs[i] == pytest.approx(true_coefs[i], abs=error_margins[i])

    def test_robust_norder_polynomial_fit_noise_and_outliers(self) -> None:

        np.random.seed(42)

        # Define x vector
        x = np.linspace(1, 10, 1000)
        # Define an exact polynomial
        true_coefs = [-100, 5, 3, 2]
        y = np.polyval(np.flip(true_coefs), x).astype(np.float32)
        # Add some noise on top
        y += np.random.normal(loc=0, scale=3, size=1000)
        # Add some outliers
        y[50:75] = 0.0
        y[900:925] = 1000.0

        # Run with the "Linear" estimator
        coefs, deg = xdem.fit.robust_norder_polynomial_fit(
            x, y, estimator_name="Linear", linear_pkg="scipy", loss="soft_l1", method="trf", f_scale=0.5
        )

        # TODO: understand why this is not robust since moving from least_squares() to curve_fit(), while the
        #  arguments passed are exactly the same...

        # Scipy solution should be quite robust to outliers/noise (with the soft_l1 method and f_scale parameter)
        # However, it is subject to random processes inside the scipy function (couldn't find how to fix those...)
        # It can find a degree 3, or 4 with coefficient close to 0
        # assert deg in [3, 4]
        # acceptable_scipy_linear_margins = [3, 3, 1, 1]
        # for i in range(4):
        #     assert coefs[i] == pytest.approx(true_coefs[i], abs=acceptable_scipy_linear_margins[i])

        # The sklearn Linear solution with MSE cost function will not be robust
        coefs2, deg2 = xdem.fit.robust_norder_polynomial_fit(
            x, y, estimator_name="Linear", linear_pkg="sklearn", cost_func=mean_squared_error, margin_improvement=50
        )
        # It won't find the right degree because of the outliers and noise
        assert deg2 != 3
        # Using the median absolute error should improve the fit
        coefs3, deg3 = xdem.fit.robust_norder_polynomial_fit(
            x, y, estimator_name="Linear", linear_pkg="sklearn", cost_func=median_absolute_error, margin_improvement=50
        )
        # Will find the right degree, but won't find the right coefficients because of the outliers and noise
        assert deg3 == 3
        sklearn_linear_error = [50, 10, 5, 0.5]
        for i in range(4):
            assert np.abs(coefs3[i] - true_coefs[i]) > sklearn_linear_error[i]

        # Now, the robust estimators
        # Theil-Sen should have better coefficients
        coefs4, deg4 = xdem.fit.robust_norder_polynomial_fit(x, y, estimator_name="Theil-Sen", random_state=42)
        assert deg4 == 3
        # High degree coefficients should be well constrained
        assert coefs4[2] == pytest.approx(true_coefs[2], abs=1)
        assert coefs4[3] == pytest.approx(true_coefs[3], abs=1)

        # RANSAC also works
        coefs5, deg5 = xdem.fit.robust_norder_polynomial_fit(x, y, estimator_name="RANSAC", random_state=42)
        assert deg5 == 3

        # Huber should perform well, close to the scipy robust solution
        coefs6, deg6 = xdem.fit.robust_norder_polynomial_fit(x, y, estimator_name="Huber")
        assert deg6 == 3
        for i in range(3):
            assert coefs6[i + 1] == pytest.approx(true_coefs[i + 1], abs=1)

    def test_robust_nfreq_sumsin_fit(self) -> None:

        # Define X vector
        x = np.linspace(0, 10, 1000)
        # Define exact sum of sinusoid signal
        true_coefs = np.array([(5, 3, np.pi), (2, 0.5, 0)]).flatten()
        y = xdem.fit.sumsin_1d(x, *true_coefs)

        # Check that the function runs (we passed a small niter to reduce the computing time of the test)
        coefs, deg = xdem.fit.robust_nfreq_sumsin_fit(x, y, random_state=42, niter=40)

        # Check that the estimated sum of sinusoid correspond to the input, with better tolerance on the highest
        # amplitude sinusoid
        # TODO: Work on making results not random between OS with basinhopping, this currently fails on Windows and Mac
        if platform.system() == "Linux":
            # Test all parameters
            for i in np.arange(6):
                # For the phase, check the circular variable with distance to modulo 2 pi
                if (i + 1) % 3 == 0:
                    coef_diff = coefs[i] - true_coefs[i] % (2 * np.pi)
                    assert np.minimum(coef_diff, np.abs(2 * np.pi - coef_diff)) < 0.1
                # Else check normally
                else:
                    assert coefs[i] == pytest.approx(true_coefs[i], abs=0.1)

        # Check that using custom arguments does not trigger an error
        bounds = [(1, 7), (1, 10), (0, 2 * np.pi), (1, 7), (0.1, 4), (0, 2 * np.pi)]
        coefs, deg = xdem.fit.robust_nfreq_sumsin_fit(
            x, y, bounds_amp_wave_phase=bounds, max_nb_frequency=2, hop_length=0.01, random_state=42, niter=1
        )

    def test_robust_nfreq_simsin_fit_noise_and_outliers(self) -> None:

        # Check robustness to outliers
        np.random.seed(42)
        # Define X vector
        x = np.linspace(0, 10, 1000)
        # Define exact sum of sinusoid signal
        true_coefs = np.array([(5, 3, np.pi), (3, 0.5, 0)]).flatten()
        y = xdem.fit.sumsin_1d(x, *true_coefs)

        # Add some noise
        y += np.random.normal(loc=0, scale=0.25, size=1000)
        # Add some outliers
        y[50:75] = -10
        y[900:925] = 10

        # Define first guess for bounds and run
        bounds = [(3, 7), (1, 5), (0, 2 * np.pi), (1, 7), (0.1, 1), (0, 2 * np.pi), (0.1, 1), (0.1, 1), (0, 2 * np.pi)]
        coefs, deg = xdem.fit.robust_nfreq_sumsin_fit(x, y, random_state=42, bounds_amp_wave_phase=bounds, niter=5)

        # Should be less precise, but still on point
        # We need to re-order output coefficient to match input
        if coefs[3] > coefs[0]:
            coefs = np.concatenate((coefs[3:], coefs[0:3]))

        # Check values
        for i in range(2):
            assert coefs[3 * i] == pytest.approx(true_coefs[3 * i], abs=0.2)
            assert coefs[3 * i + 1] == pytest.approx(true_coefs[3 * i + 1], abs=0.2)
            error_phase = min(
                np.abs(coefs[3 * i + 2] - true_coefs[3 * i + 2]),
                np.abs(2 * np.pi - (coefs[3 * i + 2] - true_coefs[3 * i + 2])),
            )
            assert error_phase < 0.2



"""Test the xdem.misc functions."""
from __future__ import annotations

import os
import warnings

import pytest
import yaml  # type: ignore
from packaging.version import Version

import xdem
import xdem.misc


class TestMisc:
    def test_environment_files(self) -> None:
        """Check that environment yml files are properly written: all dependencies of env are also in dev-env"""

        fn_env = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "environment.yml"))
        fn_devenv = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "dev-environment.yml"))

        # Load the yml as dictionaries
        yaml_env = yaml.safe_load(open(fn_env))
        yaml_devenv = yaml.safe_load(open(fn_devenv))

        # Extract the dependencies values
        conda_dep_env = yaml_env["dependencies"]
        conda_dep_devenv = yaml_devenv["dependencies"]

        # Check if there is any pip dependency, if yes pop it from the end of the list
        if isinstance(conda_dep_devenv[-1], dict):
            pip_dep_devenv = conda_dep_devenv.pop()

            # Check if there is a pip dependency in the normal env as well, if yes pop it also
            if isinstance(conda_dep_env[-1], dict):
                pip_dep_env = conda_dep_env.pop()

                # The diff below computes the dependencies that are in env but not in dev-env
                # It should be empty, otherwise we raise an error
                diff_pip_check = list(set(pip_dep_env) - set(pip_dep_devenv))
                assert len(diff_pip_check) == 0

        # We do the same for the conda dependency, first a sanity check that everything that is in env is also in dev-ev
        diff_conda_check = list(set(conda_dep_env) - set(conda_dep_devenv))
        assert len(diff_conda_check) == 0

    @pytest.mark.parametrize("deprecation_increment", [-1, 0, 1, None])  # type: ignore
    @pytest.mark.parametrize("details", [None, "It was completely useless!", "dunnowhy"])  # type: ignore
    def test_deprecate(self, deprecation_increment: int | None, details: str | None) -> None:
        """
        Test the deprecation warnings/errors.

        If the removal_version is larger than the current, it should warn.
        If the removal_version is smaller or equal, it should raise an error.

        :param deprecation_increment: The version number relative to the current version.
        :param details: An optional explanation for the description.
        """

        warnings.simplefilter("error")

        current_version = xdem.version.version

        # Set the removal version to be the current version plus the increment (e.g. 0.0.5 + 1 -> 0.0.6)
        removal_version = (
            current_version[:-1] + str(int(current_version.rsplit(".", 1)[1]) + deprecation_increment)
            if deprecation_increment is not None
            else None
        )

        # Define a function with no use that is marked as deprecated.
        @xdem.misc.deprecate(removal_version, details=details)  # type: ignore
        def useless_func() -> int:
            return 1

        # Example of why Version needs to be used below
        assert not "0.0.10" > "0.0.8"
        assert Version("0.0.10") > Version("0.0.8")

        # If True, a warning is expected. If False, a ValueError is expected.
        should_warn = removal_version is None or Version(removal_version) > Version(current_version)

        # Add the expected text depending on the parametrization.
        text = (
            "Call to deprecated function 'useless_func'."
            if should_warn
            else f"Deprecated function 'useless_func' was removed in {removal_version}."
        )

        if details is not None:
            text += " " + details.strip().capitalize()

            if not any(text.endswith(c) for c in ".!?"):
                text += "."

        if should_warn and removal_version is not None:
            text += f" This functionality will be removed in version {removal_version}."
        elif not should_warn:
            text += f" Current version: {current_version}."

        # Expect either a warning or an exception with the right text.
        if should_warn:
            with pytest.warns(DeprecationWarning, match="^" + text + "$"):
                useless_func()
        else:
            with pytest.raises(ValueError, match="^" + text + "$"):
                useless_func()

    def test_diff_environment_yml(self, capsys) -> None:  # type: ignore

        # Test with synthetic environment
        env = {"dependencies": ["python==3.9", "numpy", "fiona"]}
        devenv = {"dependencies": ["python==3.9", "numpy", "fiona", "opencv"]}

        # This should print the difference between the two
        xdem.misc.diff_environment_yml(env, devenv, input_dict=True, print_dep="conda")

        # Capture the stdout and check it is indeed the right diff
        captured = capsys.readouterr().out
        assert captured == "opencv\n"

        # This should print the difference including pip
        xdem.misc.diff_environment_yml(env, devenv, input_dict=True, print_dep="both")

        captured = capsys.readouterr().out
        assert captured == "opencv\nNone\n"

        env2 = {"dependencies": ["python==3.9", "numpy", "fiona"]}
        devenv2 = {"dependencies": ["python==3.9", "numpy", "fiona", "opencv", {"pip": ["geoutils", "-e ./"]}]}

        # The diff function should not account for -e ./ that is the local install for developers
        xdem.misc.diff_environment_yml(env2, devenv2, input_dict=True, print_dep="both")
        captured = capsys.readouterr().out

        assert captured == "opencv\ngeoutils\n"

        # This should print only pip
        xdem.misc.diff_environment_yml(env2, devenv2, input_dict=True, print_dep="pip")
        captured = capsys.readouterr().out

        assert captured == "geoutils\n"

        # This should raise an error because print_dep is not well defined
        with pytest.raises(ValueError, match='The argument "print_dep" can only be "conda", "pip" or "both".'):
            xdem.misc.diff_environment_yml(env2, devenv2, input_dict=True, print_dep="lol")

        # When the dependencies are not defined in dev-env but in env, it should raise an error
        # For normal dependencies
        env3 = {"dependencies": ["python==3.9", "numpy", "fiona", "lol"]}
        devenv3 = {"dependencies": ["python==3.9", "numpy", "fiona", "opencv", {"pip": ["geoutils"]}]}
        with pytest.raises(ValueError, match="The following dependencies are listed in env but not dev-env: lol"):
            xdem.misc.diff_environment_yml(env3, devenv3, input_dict=True, print_dep="pip")

        # For pip dependencies
        env4 = {"dependencies": ["python==3.9", "numpy", "fiona", {"pip": ["lol"]}]}
        devenv4 = {"dependencies": ["python==3.9", "numpy", "fiona", "opencv", {"pip": ["geoutils"]}]}
        with pytest.raises(ValueError, match="The following pip dependencies are listed in env but not dev-env: lol"):
            xdem.misc.diff_environment_yml(env4, devenv4, input_dict=True, print_dep="pip")



"""Routines for vertical CRS transformation (fully based on pyproj)."""
from __future__ import annotations

import http.client
import os
import pathlib
import warnings
from typing import Literal, TypedDict

import pyproj
from pyproj import CRS
from pyproj.crs import BoundCRS, CompoundCRS, GeographicCRS, VerticalCRS
from pyproj.crs.coordinate_system import Ellipsoidal3DCS
from pyproj.crs.enums import Ellipsoidal3DCSAxis
from pyproj.transformer import TransformerGroup

from xdem._typing import MArrayf, NDArrayf

# Sources for defining vertical references:
# AW3D30: https://www.eorc.jaxa.jp/ALOS/en/aw3d30/aw3d30v11_format_e.pdf
# SRTMGL1: https://lpdaac.usgs.gov/documents/179/SRTM_User_Guide_V3.pdf
# SRTMv4.1: http://www.cgiar-csi.org/data/srtm-90m-digital-elevation-database-v4-1
# ASTGTM2/ASTGTM3: https://lpdaac.usgs.gov/documents/434/ASTGTM_User_Guide_V3.pdf
# NASADEM: https://lpdaac.usgs.gov/documents/592/NASADEM_User_Guide_V1.pdf, HGTS is ellipsoid, HGT is EGM96 geoid !!
# ArcticDEM (mosaic and strips): https://www.pgc.umn.edu/data/arcticdem/
# REMA (mosaic and strips): https://www.pgc.umn.edu/data/rema/
# TanDEM-X 90m global: https://geoservice.dlr.de/web/dataguide/tdm90/
# COPERNICUS DEM: https://spacedata.copernicus.eu/web/cscda/dataset-details?articleId=394198
vcrs_dem_products = {
    "ArcticDEM/REMA/EarthDEM": "Ellipsoid",
    "TDM1": "Ellipsoid",
    "NASADEM-HGTS": "Ellipsoid",
    "AW3D30": "EGM96",
    "SRTMv4.1": "EGM96",
    "ASTGTM2": "EGM96",
    "ASTGTM3": "EGM96",
    "NASADEM-HGT": "EGM96",
    "COPDEM": "EGM08",
}


def _parse_vcrs_name_from_product(product: str) -> str | None:
    """
    Parse vertical CRS name from DEM product name.

    :param product: Product name (typically from satimg.parse_metadata_from_fn).

    :return: vcrs_name: Vertical CRS name.
    """

    if product in vcrs_dem_products.keys():
        vcrs_name = vcrs_dem_products[product]
    else:
        vcrs_name = None

    return vcrs_name


def _build_ccrs_from_crs_and_vcrs(crs: CRS, vcrs: CRS | Literal["Ellipsoid"]) -> CompoundCRS | CRS:
    """
    Build a compound CRS from a horizontal CRS and a vertical CRS.

    :param crs: Horizontal CRS.
    :param vcrs: Vertical CRS.

    :return: Compound CRS (horizontal + vertical).
    """

    # If a vertical CRS was passed, build a compound CRS with horizontal + vertical
    # This requires transforming the horizontal CRS to 2D in case it was 3D
    # Using CRS() because rasterio.CRS does not allow to call .name otherwise...
    if isinstance(vcrs, CRS):
        # If pyproj >= 3.5.1, we can use CRS.to_2d()
        from packaging.version import Version

        if Version(pyproj.__version__) > Version("3.5.0"):
            crs_from = CRS(crs).to_2d()
            ccrs = CompoundCRS(
                name="Horizontal: " + CRS(crs).name + "; Vertical: " + vcrs.name,
                components=[crs_from, vcrs],
            )
        # Otherwise, we have to raise an error if the horizontal CRS is already 3D
        else:
            crs_from = CRS(crs)
            # If 3D
            if len(crs_from.axis_info) > 2:
                raise NotImplementedError(
                    "pyproj >= 3.5.1 is required to demote a 3D CRS to 2D and be able to compound "
                    "with a new vertical CRS. Update your dependencies or pass the 2D source CRS "
                    "manually."
                )
            # If 2D
            else:
                ccrs = CompoundCRS(
                    name="Horizontal: " + CRS(crs).name + "; Vertical: " + vcrs.name,
                    components=[crs_from, vcrs],
                )

    # Else if "Ellipsoid" was passed, there is no vertical reference
    # We still have to return the CRS in 3D
    elif isinstance(vcrs, str) and vcrs.lower() == "ellipsoid":
        ccrs = CRS(crs).to_3d()
    else:
        raise ValueError("Invalid vcrs given. Must be a vertical CRS or the literal string 'Ellipsoid'.")

    return ccrs


def _build_vcrs_from_grid(grid: str, old_way: bool = False) -> CompoundCRS:
    """
    Build a compound CRS from a vertical CRS grid path.

    :param grid: Path to grid for vertical reference.
    :param old_way: Whether to use the new or old way of building the compound CRS with pyproj (for testing purposes).

    :return: Compound CRS (horizontal + vertical).
    """

    if not os.path.exists(os.path.join(pyproj.datadir.get_data_dir(), grid)):
        warnings.warn(
            "Grid not found in "
            + str(pyproj.datadir.get_data_dir())
            + ". Attempting to download from https://cdn.proj.org/..."
        )
        from pyproj.sync import _download_resource_file

        try:
            _download_resource_file(
                file_url=os.path.join("https://cdn.proj.org/", grid),
                short_name=grid,
                directory=pyproj.datadir.get_data_dir(),
                verbose=False,
            )
        except http.client.InvalidURL:
            raise ValueError(
                "The provided grid '{}' does not exist at https://cdn.proj.org/. "
                "Provide an existing grid.".format(grid)
            )

    # The old way: see https://gis.stackexchange.com/questions/352277/.
    if old_way:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", module="pyproj")
            ccrs = pyproj.Proj(init="EPSG:4326", geoidgrids=grid).crs
            bound_crs = ccrs.sub_crs_list[1]

    # The clean way
    else:
        # First, we build a bounds CRS (the vertical CRS relative to geographic)
        vertical_crs = VerticalCRS(
            name="unknown using geoidgrids=" + grid, datum='VDATUM["unknown using geoidgrids=' + grid + '"]'
        )
        geographic3d_crs = GeographicCRS(
            name="WGS 84",
            ellipsoidal_cs=Ellipsoidal3DCS(axis=Ellipsoidal3DCSAxis.LATITUDE_LONGITUDE_HEIGHT),
        )
        bound_crs = BoundCRS(
            source_crs=vertical_crs,
            target_crs=geographic3d_crs,
            transformation={
                "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
                "type": "Transformation",
                "name": "unknown to WGS84 ellipsoidal height",
                "source_crs": vertical_crs.to_json_dict(),
                "target_crs": geographic3d_crs.to_json_dict(),
                "method": {"name": "GravityRelatedHeight to Geographic3D"},
                "parameters": [
                    {
                        "name": "Geoid (height correction) model file",
                        "value": grid,
                        "id": {"authority": "EPSG", "code": 8666},
                    }
                ],
            },
        )

    return bound_crs


# Define types of common Vertical CRS dictionary
class VCRSMetaDict(TypedDict, total=False):
    grid: str
    epsg: int


_vcrs_meta: dict[str, VCRSMetaDict] = {
    "EGM08": {"grid": "us_nga_egm08_25.tif", "epsg": 3855},  # EGM2008 at 2.5 minute resolution
    "EGM96": {"grid": "us_nga_egm96_15.tif", "epsg": 5773},  # EGM1996 at 15 minute resolution
}


def _vcrs_from_crs(crs: CRS) -> CRS:
    """Get the vertical CRS from a CRS."""

    # Check if CRS is 3D
    if len(crs.axis_info) > 2:

        # Check if CRS has a vertical compound
        if any(subcrs.is_vertical for subcrs in crs.sub_crs_list):
            # Then we get the first vertical CRS (should be only one anyway)
            vcrs = [subcrs for subcrs in crs.sub_crs_list if subcrs.is_vertical][0]
        # Otherwise, it's a 3D CRS based on an ellipsoid
        else:
            vcrs = "Ellipsoid"
    # Otherwise, the CRS is 2D and there is no vertical CRS
    else:
        vcrs = None

    return vcrs


def _vcrs_from_user_input(
    vcrs_input: Literal["Ellipsoid"] | Literal["EGM08"] | Literal["EGM96"] | str | pathlib.Path | CRS | int,
) -> VerticalCRS | BoundCRS | Literal["Ellipsoid"]:
    """
    Parse vertical CRS from user input.

    :param vcrs_input: Vertical coordinate reference system either as a name ("Ellipsoid", "EGM08", "EGM96"),
        an EPSG code or pyproj.crs.VerticalCRS, or a path to a PROJ grid file (https://github.com/OSGeo/PROJ-data).

    :return: Vertical CRS.
    """

    # Raise errors if input type is wrong (allow CRS instead of VerticalCRS for broader error messages below)
    if not isinstance(vcrs_input, (str, pathlib.Path, CRS, int)):
        raise TypeError(f"New vertical CRS must be a string, path or VerticalCRS, received {type(vcrs_input)}.")

    # If input is ellipsoid
    if (
        (isinstance(vcrs_input, str) and (vcrs_input.lower() == "ellipsoid" or vcrs_input.upper() == "WGS84"))
        or (isinstance(vcrs_input, int) and vcrs_input in [4326, 4979])
        or (isinstance(vcrs_input, CRS) and vcrs_input.to_epsg() in [4326, 4979])
    ):
        return "Ellipsoid"

    # Define CRS in case EPSG or CRS was passed
    if isinstance(vcrs_input, (int, CRS)):
        if isinstance(vcrs_input, int):
            vcrs = CRS.from_epsg(vcrs_input)
        else:
            vcrs = vcrs_input

        # Raise errors if the CRS constructed is not vertical or has other components
        if isinstance(vcrs, CRS) and not vcrs.is_vertical:
            raise ValueError(
                "New vertical CRS must have a vertical axis, '{}' does not "
                "(check with `CRS.is_vertical`).".format(vcrs.name)
            )
        elif isinstance(vcrs, CRS) and vcrs.is_vertical and len(vcrs.axis_info) > 2:
            warnings.warn(
                "New vertical CRS has a vertical dimension but also other components, "
                "extracting the vertical reference only."
            )
            vcrs = _vcrs_from_crs(vcrs)

    # If a string was passed
    else:
        # If a name is passed, define CRS based on dict
        if isinstance(vcrs_input, str) and vcrs_input.upper() in _vcrs_meta.keys():
            vcrs_meta = _vcrs_meta[vcrs_input]
            vcrs = CRS.from_epsg(vcrs_meta["epsg"])
        # Otherwise, attempt to read a grid from the string
        else:
            if isinstance(vcrs_input, pathlib.Path):
                grid = vcrs_input.name
            else:
                grid = vcrs_input
            vcrs = _build_vcrs_from_grid(grid=grid)

    return vcrs


def _grid_from_user_input(vcrs_input: str | pathlib.Path | int | CRS) -> str | None:

    # If a grid or name was passed, get grid name
    if isinstance(vcrs_input, (str, pathlib.Path)):
        # If the string is within the supported names
        if isinstance(vcrs_input, str) and vcrs_input in _vcrs_meta.keys():
            grid = _vcrs_meta[vcrs_input]["grid"]
        # If it's a pathlib path
        elif isinstance(vcrs_input, pathlib.Path):
            grid = vcrs_input.name
        # Or an ellipsoid
        elif vcrs_input.lower() == "ellipsoid":
            grid = None
        # Or a string path
        else:
            grid = vcrs_input
    # Otherwise, there is none
    else:
        grid = None

    return grid


def _transform_zz(
    crs_from: CRS, crs_to: CRS, xx: NDArrayf, yy: NDArrayf, zz: MArrayf | NDArrayf | int | float
) -> MArrayf | NDArrayf | int | float:
    """
    Transform elevation to a new 3D CRS.

    :param crs_from: Source CRS.
    :param crs_to: Destination CRS.
    :param xx: X coordinates.
    :param yy: Y coordinates.
    :param zz: Z coordinates.

    :return: Transformed Z coordinates.
    """

    # Find all possible transforms
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "Best transformation is not available")
        trans_group = TransformerGroup(crs_from=crs_from, crs_to=crs_to, always_xy=True)

    # Download grid if best available is not on disk, download and re-initiate the object
    if not trans_group.best_available:
        trans_group.download_grids()
        trans_group = TransformerGroup(crs_from=crs_from, crs_to=crs_to, always_xy=True)

    # If the best available grid is still not there, raise a warning
    if not trans_group.best_available:
        warnings.warn(
            category=UserWarning,
            message="Best available grid for transformation could not be downloaded, "
            "applying the next best available (caution: might apply no transform at all).",
        )
    transformer = trans_group.transformers[0]

    # Will preserve the mask of the masked-array since pyproj 3.4
    zz_trans = transformer.transform(xx, yy, zz)[2]

    return zz_trans



from __future__ import annotations

import sys
from typing import Any

import numpy as np

# Only for Python >= 3.9
if sys.version_info.minor >= 9:

    from numpy.typing import NDArray  # this syntax works starting on Python 3.9

    NDArrayf = NDArray[np.floating[Any]]
    MArrayf = np.ma.masked_array[Any, np.dtype[np.floating[Any]]]

else:
    NDArrayf = np.array  # type: ignore
    MArrayf = np.ma.masked_array  # type: ignore



"""Functions to test the difference of DEMs tools."""
import warnings

import geoutils as gu
import numpy as np

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import xdem


class TestdDEM:
    dem_2009 = xdem.DEM(xdem.examples.get_path("longyearbyen_ref_dem"))
    dem_1990 = xdem.DEM(xdem.examples.get_path("longyearbyen_tba_dem"))
    outlines_1990 = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))

    ddem = xdem.dDEM(dem_2009 - dem_1990, start_time=np.datetime64("1990-08-01"), end_time=np.datetime64("2009-08-01"))

    def test_init(self) -> None:
        """Test that the dDEM object was instantiated correctly."""
        assert isinstance(self.ddem, xdem.dDEM)
        assert isinstance(self.ddem.data, np.ma.masked_array)

    def test_copy(self) -> None:
        """Test that copying works as it should."""
        ddem2 = self.ddem.copy()

        assert isinstance(ddem2, xdem.dDEM)

        ddem2.data += 1

        assert not self.ddem.raster_equal(ddem2)

    def test_filled_data(self) -> None:
        """Test that the filled_data property points to the right data."""
        ddem2 = self.ddem.copy()

        assert not np.any(np.isnan(ddem2.data)) or np.all(~ddem2.data.mask)
        assert ddem2.filled_data is not None

        assert np.count_nonzero(np.isnan(ddem2.data)) == 0
        ddem2.data.ravel()[0] = np.nan

        assert np.count_nonzero(np.isnan(ddem2.data)) == 1

        assert ddem2.filled_data is None

        ddem2.interpolate(method="linear")

        assert ddem2.fill_method is not None

    def test_regional_hypso(self) -> None:
        """Test the regional hypsometric approach."""
        ddem = self.ddem.copy()
        ddem.data.mask = np.zeros_like(ddem.data, dtype=bool)
        ddem.data.mask.ravel()[np.random.choice(ddem.data.size, 50000, replace=False)] = True
        assert np.count_nonzero(ddem.data.mask) > 0

        assert ddem.filled_data is None

        ddem.interpolate(method="regional_hypsometric", reference_elevation=self.dem_2009, mask=self.outlines_1990)

        assert ddem._filled_data is not None
        assert isinstance(ddem.filled_data, np.ndarray)

        assert ddem.filled_data.shape == ddem.data.shape

        assert np.abs(np.nanmean(self.ddem.data - ddem.filled_data)) < 1

    def test_local_hypso(self) -> None:
        """Test the local hypsometric approach."""
        ddem = self.ddem.copy()
        scott_1990 = self.outlines_1990.query("NAME == 'Scott Turnerbreen'")
        ddem.data.mask = np.zeros_like(ddem.data, dtype=bool)
        ddem.data.mask.ravel()[np.random.choice(ddem.data.size, 50000, replace=False)] = True
        assert np.count_nonzero(ddem.data.mask) > 0

        assert ddem.filled_data is None

        ddem.interpolate(method="local_hypsometric", reference_elevation=self.dem_2009.data, mask=scott_1990)
        assert np.abs(np.mean(self.ddem.data - ddem.filled_data)) < 1



"""Functions to test the example data."""
from __future__ import annotations

import geoutils as gu
import numpy as np
import pytest
from geoutils import Raster, Vector

from xdem import examples
from xdem._typing import NDArrayf


def load_examples() -> tuple[Raster, Raster, Vector, Raster]:
    """Load example files to try coregistration methods with."""

    ref_dem = Raster(examples.get_path("longyearbyen_ref_dem"))
    tba_dem = Raster(examples.get_path("longyearbyen_tba_dem"))
    glacier_mask = Vector(examples.get_path("longyearbyen_glacier_outlines"))
    ddem = Raster(examples.get_path("longyearbyen_ddem"))

    return ref_dem, tba_dem, glacier_mask, ddem


class TestExamples:

    ref_dem, tba_dem, glacier_mask, ddem = load_examples()

    @pytest.mark.parametrize(
        "rst_and_truevals",
        [
            (ref_dem, np.array([868.6489, 623.42194, 180.57921, 267.30765, 601.67615], dtype=np.float32)),
            (tba_dem, np.array([875.2358, 625.0544, 182.9936, 272.6586, 606.2897], dtype=np.float32)),
            (
                ddem,
                np.array(
                    [
                        -4.669189453125000000e-02,
                        -7.413940429687500000e-01,
                        1.499481201171875000e-01,
                        1.095550537109375000e00,
                        -5.904846191406250000e00,
                    ],
                    dtype=np.float32,
                ),
            ),
        ],
    )  # type: ignore
    def test_array_content(self, rst_and_truevals: tuple[Raster, NDArrayf]) -> None:
        """Let's ensure the data arrays in the examples are always the same by checking randomly some values"""

        rst = rst_and_truevals[0]
        truevals = rst_and_truevals[1]
        np.random.seed(42)
        values = np.random.choice(rst.data.data.flatten(), size=5, replace=False)

        assert values == pytest.approx(truevals)

    # Note: Following PR #329, no gaps on DEM edges after coregistration
    @pytest.mark.parametrize("rst_and_truenodata", [(ref_dem, 0), (tba_dem, 0), (ddem, 0)])  # type: ignore
    def test_array_nodata(self, rst_and_truenodata: tuple[Raster, int]) -> None:
        """Let's also check that the data arrays have always the same number of not finite values"""

        rst = rst_and_truenodata[0]
        truenodata = rst_and_truenodata[1]
        mask = gu.raster.get_array_and_mask(rst)[1]

        assert np.sum(mask) == truenodata



"""Filters to remove outliers and reduce noise in DEMs."""
from __future__ import annotations

import warnings

try:
    import cv2

    _has_cv2 = True
except ImportError:
    _has_cv2 = False
import numpy as np
import scipy

from xdem._typing import NDArrayf


def gaussian_filter_scipy(array: NDArrayf, sigma: float) -> NDArrayf:
    """
    Apply a Gaussian filter to a raster that may contain NaNs, using scipy's implementation.
    gaussian_filter_cv is recommended as it is usually faster, but this depends on the value of sigma.

    N.B: kernel_size is set automatically based on sigma.

    :param array: the input array to be filtered.
    :param sigma: the sigma of the Gaussian kernel

    :returns: the filtered array (same shape as input)
    """
    # Check that array dimension is 2 or 3
    if np.ndim(array) not in [2, 3]:
        raise ValueError(f"Invalid array shape given: {array.shape}. Expected 2D or 3D array")

    # In case array does not contain NaNs, use scipy's gaussian filter directly
    if np.count_nonzero(np.isnan(array)) == 0:
        return scipy.ndimage.gaussian_filter(array, sigma=sigma)

    # If array contain NaNs, need a more sophisticated approach
    # Inspired by https://stackoverflow.com/a/36307291
    else:

        # Run filter on a copy with NaNs set to 0
        array_no_nan = array.copy()
        array_no_nan[np.isnan(array)] = 0
        gauss_no_nan = scipy.ndimage.gaussian_filter(array_no_nan, sigma=sigma)
        del array_no_nan

        # Mask of NaN values
        nan_mask = 0 * array.copy() + 1
        nan_mask[np.isnan(array)] = 0
        gauss_mask = scipy.ndimage.gaussian_filter(nan_mask, sigma=sigma)
        del nan_mask

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="invalid value encountered")
            gauss = gauss_no_nan / gauss_mask

        return gauss


def gaussian_filter_cv(array: NDArrayf, sigma: float) -> NDArrayf:
    """
    Apply a Gaussian filter to a raster that may contain NaNs, using OpenCV's implementation.
    Arguments are for now hard-coded to be identical to scipy.

    N.B: kernel_size is set automatically based on sigma

    :param array: the input array to be filtered.
    :param sigma: the sigma of the Gaussian kernel

    :returns: the filtered array (same shape as input)
    """
    if not _has_cv2:
        raise ValueError("Optional dependency needed. Install 'opencv'")

    # Check that array dimension is 2, or can be squeezed to 2D
    orig_shape = array.shape
    if len(orig_shape) == 2:
        pass
    elif len(orig_shape) == 3:
        if orig_shape[0] == 1:
            array = array.squeeze()
        else:
            raise NotImplementedError("Case of array of dimension 3 not implemented")
    else:
        raise ValueError(f"Invalid array shape given: {orig_shape}. Expected 2D or 3D array")

    # In case array does not contain NaNs, use OpenCV's gaussian filter directly
    # With kernel size (0, 0), i.e. set to default, and borderType=BORDER_REFLECT, the output is equivalent to scipy
    if np.count_nonzero(np.isnan(array)) == 0:
        gauss = cv2.GaussianBlur(array, (0, 0), sigmaX=sigma, borderType=cv2.BORDER_REFLECT)

    # If array contain NaNs, need a more sophisticated approach
    # Inspired by https://stackoverflow.com/a/36307291
    else:

        # Run filter on a copy with NaNs set to 0
        array_no_nan = array.copy()
        array_no_nan[np.isnan(array)] = 0
        gauss_no_nan = cv2.GaussianBlur(array_no_nan, (0, 0), sigmaX=sigma, borderType=cv2.BORDER_REFLECT)
        del array_no_nan

        # Mask of NaN values
        nan_mask = 0 * array.copy() + 1
        nan_mask[np.isnan(array)] = 0
        gauss_mask = cv2.GaussianBlur(nan_mask, (0, 0), sigmaX=sigma, borderType=cv2.BORDER_REFLECT)
        del nan_mask

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="invalid value encountered")
            gauss = gauss_no_nan / gauss_mask

    return gauss.reshape(orig_shape)


# Median filters

# To be added

# Min/max filters

# To be added


def distance_filter(array: NDArrayf, radius: float, outlier_threshold: float) -> NDArrayf:
    """
    Filter out pixels whose value is distant more than a set threshold from the average value of all neighbor \
pixels within a given radius.
    Filtered pixels are set to NaN.

    TO DO: Add an option on how the "average" value should be calculated, i.e. using a Gaussian, median etc filter.

    :param array: the input array to be filtered.
    :param radius: the radius in which the average value is calculated (for Gaussian filter, this is sigma).
    :param outlier_threshold: the minimum difference abs(array - mean) for a pixel to be considered an outlier.

    :returns: the filtered array (same shape as input)
    """
    # Calculate the average value within the radius
    smooth = gaussian_filter_cv(array, sigma=radius)

    # Filter outliers
    outliers = (np.abs(array - smooth)) > outlier_threshold
    out_array = np.copy(array)
    out_array[outliers] = np.nan

    return out_array



"""Functions to test the documentation."""
import os
import platform
import shutil
import warnings

import sphinx.cmd.build


class TestDocs:
    docs_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "../", "doc/")
    n_threads = os.getenv("N_CPUS")

    def test_example_code(self) -> None:
        """Try running each python script in the doc/source/code\
                directory and check that it doesn't raise an error."""
        current_dir = os.getcwd()
        os.chdir(os.path.join(self.docs_dir, "source"))

        def run_code(filename: str) -> None:
            """Run a python script in one thread."""
            with open(filename) as infile:
                # Run everything except plt.show() calls.
                with warnings.catch_warnings():
                    # When running the code asynchronously, matplotlib complains a bit
                    ignored_warnings = [
                        "Starting a Matplotlib GUI outside of the main thread",
                        ".*fetching the attribute.*Polygon.*",
                    ]
                    # This is a GeoPandas issue
                    warnings.simplefilter("error")

                    for warning_text in ignored_warnings:
                        warnings.filterwarnings("ignore", warning_text)
                    try:
                        exec(infile.read().replace("plt.show()", "plt.close()"))
                    except Exception as exception:
                        if isinstance(exception, DeprecationWarning):
                            print(exception)
                        else:
                            raise RuntimeError(f"Failed on {filename}") from exception

        filenames = [os.path.join("code", filename) for filename in os.listdir("code/") if filename.endswith(".py")]

        for filename in filenames:
            run_code(filename)
        """
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=int(self.n_threads) if self.n_threads is not None else None
        ) as executor:
            list(executor.map(run_code, filenames))
        """

        os.chdir(current_dir)

    def test_build(self) -> None:
        """Try building the doc and see if it works."""

        # Test only on Linux
        if platform.system() == "Linux":
            # Remove the build directory if it exists.
            if os.path.isdir(os.path.join(self.docs_dir, "build")):
                shutil.rmtree(os.path.join(self.docs_dir, "build"))

            return_code = sphinx.cmd.build.main(
                [
                    "-j",
                    "1",
                    os.path.join(self.docs_dir, "source"),
                    os.path.join(self.docs_dir, "build", "html"),
                ]
            )

            assert return_code == 0



from __future__ import annotations

import os

from setuptools import setup

FULLVERSION = "0.0.11"
VERSION = FULLVERSION

with open(os.path.join(os.path.dirname(__file__), "README.md")) as infile:
    LONG_DESCRIPTION = infile.read()

setup(
    name="xdem",
    version=FULLVERSION,
    description="Set of tools to manipulate Digital Elevation Models (DEMs) ",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/GlacioHack/xdem",
    author="The GlacioHack Team",
    author_email="this-is-not-an-email@a.com",  # This is needed for PyPi unfortunately.
    license="BSD-3",
    packages=["xdem"],
    install_requires=[
        "numpy",
        "scipy",
        "rasterio",
        "geopandas",
        "pyproj",
        "tqdm",
        "scikit-gstat",
        "scikit-image",
        "geoutils",
    ],
    extras_require={
        "rioxarray": ["rioxarray"],
        "richdem": ["richdem"],
        "opencv": ["opencv"],
        "pytransform3d": ["pytransform3d"],
    },
    python_requires=">=3.8",
    scripts=[],
    zip_safe=False,
)

write_version = True


def write_version_py(filename: str | None = None) -> None:
    cnt = """\
version = '%s'
short_version = '%s'
"""
    if not filename:
        filename = os.path.join(os.path.dirname(__file__), "xdem", "version.py")

    a = open(filename, "w")
    try:
        a.write(cnt % (FULLVERSION, VERSION))
    finally:
        a.close()


if write_version:
    write_version_py()



"""Functions to test the volume estimation tools."""
import warnings

import geoutils as gu
import numpy as np
import pandas as pd
import pytest

import xdem


class TestLocalHypsometric:
    """Test cases for the local hypsometric method."""

    # Load example data.
    dem_2009 = gu.Raster(xdem.examples.get_path("longyearbyen_ref_dem"))
    dem_1990 = gu.Raster(xdem.examples.get_path("longyearbyen_tba_dem")).reproject(dem_2009, silent=True)
    outlines = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))
    all_outlines = outlines.copy()

    # Filter to only look at the Scott Turnerbreen glacier
    outlines.ds = outlines.ds.loc[outlines.ds["NAME"] == "Scott Turnerbreen"]
    # Create a mask where glacier areas are True
    mask = outlines.create_mask(dem_2009)

    def test_bin_ddem(self) -> None:
        """Test dDEM binning."""
        ddem = self.dem_2009 - self.dem_1990

        ddem_bins = xdem.volume.hypsometric_binning(ddem[self.mask], self.dem_2009[self.mask], bins=50, kind="fixed")

        ddem_bins_masked = xdem.volume.hypsometric_binning(
            np.ma.masked_array(ddem.data, mask=~self.mask.data.filled(False)),
            np.ma.masked_array(self.dem_2009.data, mask=~self.mask.data.filled(False)),
        )

        ddem_stds = xdem.volume.hypsometric_binning(
            ddem[self.mask], self.dem_2009[self.mask], aggregation_function=np.std
        )
        assert ddem_stds["value"].mean() < 50
        assert np.abs(np.mean(ddem_bins["value"] - ddem_bins_masked["value"])) < 0.01

    def test_interpolate_ddem_bins(self) -> pd.Series:
        """Test dDEM bin interpolation."""
        ddem = self.dem_2009 - self.dem_1990

        ddem_bins = xdem.volume.hypsometric_binning(ddem[self.mask], self.dem_2009[self.mask])

        # Simulate a missing bin
        ddem_bins.iloc[3, 0] = np.nan

        # Interpolate the bins and exclude bins with low pixel counts from the interpolation.
        interpolated_bins = xdem.volume.interpolate_hypsometric_bins(ddem_bins, count_threshold=200)

        # Check that the count column has not changed.
        assert np.array_equal(ddem_bins["count"], interpolated_bins["count"])

        # Assert that the absolute mean is somewhere between 0 and 40
        assert abs(np.mean(interpolated_bins["value"])) < 40
        assert abs(np.mean(interpolated_bins["value"])) > 0
        # Check that no nans exist.
        assert not np.any(np.isnan(interpolated_bins))

        # Return the value so that they can be used in other tests.
        return interpolated_bins

    def test_area_calculation(self) -> None:
        """Test the area calculation function."""
        ddem_bins = self.test_interpolate_ddem_bins()

        # Test the area calculation with normal parameters.
        bin_area = xdem.volume.calculate_hypsometry_area(
            ddem_bins, self.dem_2009[self.mask], pixel_size=self.dem_2009.res[0]
        )

        # Test area calculation with differing pixel x/y resolution.
        xdem.volume.calculate_hypsometry_area(
            ddem_bins, self.dem_2009[self.mask], pixel_size=(self.dem_2009.res[0], self.dem_2009.res[0] + 1)
        )

        # Add some nans to the reference DEM
        data_with_nans = self.dem_2009
        data_with_nans.data[2, 5] = np.nan

        # Make sure that the above results in the correct error.
        try:
            xdem.volume.calculate_hypsometry_area(ddem_bins, data_with_nans, pixel_size=self.dem_2009.res[0])
        except AssertionError as exception:
            if "DEM has NaNs" not in str(exception):
                raise exception

        # Try to pass an incorrect timeframe= parameter
        try:
            xdem.volume.calculate_hypsometry_area(
                ddem_bins, self.dem_2009[self.mask], pixel_size=self.dem_2009.res[0], timeframe="blergh"
            )
        except ValueError as exception:
            if "Argument 'timeframe=blergh' is invalid" not in str(exception):
                raise exception

        # Mess up the dDEM bins and see if it gives the correct error
        ddem_bins.iloc[3] = np.nan
        try:
            xdem.volume.calculate_hypsometry_area(ddem_bins, self.dem_2009[self.mask], pixel_size=self.dem_2009.res[0])
        except AssertionError as exception:
            if "cannot contain NaNs" not in str(exception):
                raise exception

        # The area of Scott Turnerbreen was around 3.4 km² in 1990, so this should be close to that number.
        assert 2e6 < bin_area.sum() < 5e6

    def test_ddem_bin_methods(self) -> None:
        """Test different dDEM binning methods."""
        ddem = self.dem_2009 - self.dem_1990

        # equal height is already tested in test_bin_ddem

        # Make a fixed amount of bins
        equal_count_bins = xdem.volume.hypsometric_binning(
            ddem[self.mask], self.dem_2009[self.mask], bins=50, kind="count"
        )
        assert equal_count_bins.shape[0] == 50

        # Make 50 bins with approximately the same area (pixel count)
        quantile_bins = xdem.volume.hypsometric_binning(
            ddem[self.mask], self.dem_2009[self.mask], bins=50, kind="quantile"
        )

        assert quantile_bins.shape[0] == 50
        # Make sure that the pixel count variation is low.
        assert quantile_bins["count"].std() < 1

        # Try to feed the previous bins as "custom" bins to the function.
        custom_bins = xdem.volume.hypsometric_binning(
            ddem[self.mask],
            self.dem_2009[self.mask],
            bins=np.r_[quantile_bins.index.left[0], quantile_bins.index.right],
            kind="custom",
        )

        assert custom_bins.shape[0] == quantile_bins.shape[0]


class TestNormHypsometric:
    dem_2009 = gu.Raster(xdem.examples.get_path("longyearbyen_ref_dem"))
    dem_1990 = gu.Raster(xdem.examples.get_path("longyearbyen_tba_dem")).reproject(dem_2009, silent=True)
    outlines = gu.Vector(xdem.examples.get_path("longyearbyen_glacier_outlines"))

    glacier_index_map = outlines.rasterize(dem_2009)
    ddem = dem_2009.data - dem_1990.data

    @pytest.mark.parametrize("n_bins", [5, 10, 20])  # type: ignore
    def test_regional_signal(self, n_bins: int) -> None:
        warnings.simplefilter("error")

        signal = xdem.volume.get_regional_hypsometric_signal(
            ddem=self.ddem, ref_dem=self.dem_2009, glacier_index_map=self.glacier_index_map, n_bins=n_bins
        )

        assert signal["w_mean"].min() >= 0.0
        assert signal["w_mean"].max() <= 1.0
        assert signal.index.right.max() <= 1.0
        assert signal.index.left.min() >= 0.0

        assert np.all(np.isfinite(signal.values))

    def test_interpolate_small(self) -> None:

        dem = np.arange(16, dtype="float32").reshape(4, 4)
        ddem = dem / 10
        glacier_index_map = np.ones_like(dem)

        signal = xdem.volume.get_regional_hypsometric_signal(
            ddem=ddem, ref_dem=dem, glacier_index_map=glacier_index_map, n_bins=10
        )

        # Make it so that only 1/4 of the values exist.
        ddem_orig = ddem.copy()
        ddem.ravel()[4:] = np.nan

        interpolated_ddem = xdem.volume.norm_regional_hypsometric_interpolation(
            voided_ddem=ddem,
            ref_dem=dem,
            glacier_index_map=glacier_index_map,
            regional_signal=signal,
            min_elevation_range=0.3,  # At least ~1/3 of the range need to exist.
        )

        # Validate that no interpolation was done, as the coverage was too small
        assert np.nansum(np.abs(ddem - interpolated_ddem)) == 0.0

        # Now try with a dDEM that has about 3/4 of the range.
        ddem = ddem_orig.copy()
        ddem.ravel()[12:] = np.nan

        interpolated_ddem = xdem.volume.norm_regional_hypsometric_interpolation(
            voided_ddem=ddem,
            ref_dem=dem,
            glacier_index_map=glacier_index_map,
            regional_signal=signal,
            min_elevation_range=0.3,
        )

        # Validate that values were interpolated within the measurement step-size
        assert np.nanmax(np.abs((interpolated_ddem - ddem_orig)[np.isnan(ddem)])) < 0.1

    def test_regional_hypsometric_interp(self) -> None:

        warnings.simplefilter("error")

        # Extract a normalized regional hypsometric signal.
        ddem = self.dem_2009 - self.dem_1990

        signal = xdem.volume.get_regional_hypsometric_signal(
            ddem=self.ddem, ref_dem=self.dem_2009, glacier_index_map=self.glacier_index_map
        )

        if False:
            import matplotlib.pyplot as plt

            plt.fill_between(
                signal.index.mid, signal["median"] - signal["std"], signal["median"] + signal["std"], label="Median±std"
            )
            plt.plot(signal.index.mid, signal["median"], color="black", linestyle=":", label="Median")
            plt.plot(signal.index.mid, signal["w_mean"], color="black", label="Weighted mean")

            plt.xlabel("Normalized elevation")
            plt.ylabel("Normalized elevation change")
            plt.legend()

            plt.show()

        # Try the normalized regional hypsometric interpolation.
        # Synthesize random nans in 80% of the data.
        ddem.data.mask.ravel()[np.random.choice(ddem.data.size, int(ddem.data.size * 0.80), replace=False)] = True
        # Fill the dDEM using the de-normalized signal.
        filled_ddem = xdem.volume.norm_regional_hypsometric_interpolation(
            voided_ddem=ddem, ref_dem=self.dem_2009, glacier_index_map=self.glacier_index_map
        )
        # Fill the dDEM using the de-normalized signal and create an idealized dDEM
        idealized_ddem = xdem.volume.norm_regional_hypsometric_interpolation(
            voided_ddem=ddem, ref_dem=self.dem_2009, glacier_index_map=self.glacier_index_map, idealized_ddem=True
        )
        assert not np.array_equal(filled_ddem, idealized_ddem)

        # Check that all glacier-values are finite
        assert np.count_nonzero(np.isnan(idealized_ddem)[self.glacier_index_map.get_nanarray() > 0]) == 0
        # Validate that the un-idealized dDEM has a higher gradient variance (more ups and downs)
        filled_gradient = np.linalg.norm(np.gradient(filled_ddem), axis=0)
        ideal_gradient = np.linalg.norm(np.gradient(idealized_ddem), axis=0)
        assert np.nanstd(filled_gradient) > np.nanstd(ideal_gradient)

        if False:
            import matplotlib.pyplot as plt

            plt.subplot(121)
            plt.imshow(filled_ddem, cmap="coolwarm_r", vmin=-10, vmax=10)
            plt.subplot(122)
            plt.imshow(idealized_ddem, cmap="coolwarm_r", vmin=-10, vmax=10)

            plt.show()

        # Extract the finite glacier values.
        changes = ddem.data.squeeze()[self.glacier_index_map.get_nanarray() > 0]
        changes = changes[np.isfinite(changes)]
        interp_changes = filled_ddem[self.glacier_index_map.get_nanarray() > 0]
        interp_changes = interp_changes[np.isfinite(interp_changes)]

        # Validate that the interpolated (20% data) means and stds are similar to the original (100% data)
        # These are commented outbecause the CI for some reason gets quite large variance. It works with lower
        # values on normal computers...
        # assert abs(changes.mean() - interp_changes.mean()) < 2
        # assert abs(changes.std() - interp_changes.std()) < 2
